from abc import ABC
from typing import Callable, Generic, Iterator, List, Optional

from fractal_specifications.generic.specification import Specification

from fractal.core.repositories import Entity, Repository


def filter_results(
    q: str,
    *,
    fields: List[str],
    data: Iterator[Entity],
    pre_processor: Optional[Callable[[str], str]] = None,
):
    return filter(
        lambda i: any(
            [
                q in (pre_processor(getattr(i, f)) if pre_processor else getattr(i, f))
                for f in fields
            ]
        ),
        data,
    )


class FilterRepositoryMixin(Repository[Entity], Generic[Entity], ABC):
    def find_filter(
        self,
        q: str,
        *,
        fields: List[str],
        specification: Specification = None,
        pre_processor: Optional[Callable[[str], str]] = None,
    ) -> Iterator[Entity]:
        if not q:
            return self.find(specification)
        return filter_results(
            q,
            fields=fields,
            data=self.find(specification),
            pre_processor=pre_processor,
        )



def test_is_healthy(sqlalchemy_test_repository):
    assert sqlalchemy_test_repository.is_healthy()


def test_add(sqlalchemy_test_repository, sqlalchemy_test_model):
    from fractal_specifications.generic.operators import EqualsSpecification

    obj = sqlalchemy_test_model("test")
    res = sqlalchemy_test_repository.add(obj)

    assert res == obj

    res = sqlalchemy_test_repository.find_one(EqualsSpecification("id", "test"))

    assert res == obj
    assert res.name == "test"


def test_find_no_spec(sqlalchemy_test_repository, sqlalchemy_test_model):
    obj1 = sqlalchemy_test_model("test1")
    obj2 = sqlalchemy_test_model("test2")
    sqlalchemy_test_repository.add(obj1)
    sqlalchemy_test_repository.add(obj2)

    res = list(sqlalchemy_test_repository.find())

    assert len(res) == 2
    assert res == [obj1, obj2]


def test_find_with_spec(sqlalchemy_test_repository, sqlalchemy_test_model):
    from fractal_specifications.generic.operators import EqualsSpecification

    obj1 = sqlalchemy_test_model("test1")
    obj2 = sqlalchemy_test_model("test2")
    sqlalchemy_test_repository.add(obj1)
    sqlalchemy_test_repository.add(obj2)

    res = list(sqlalchemy_test_repository.find(EqualsSpecification("id", "test1")))

    assert len(res) == 1
    assert res[0] == obj1


def test_find_with_spec_no_result(sqlalchemy_test_repository, sqlalchemy_test_model):
    from fractal_specifications.generic.operators import EqualsSpecification

    obj1 = sqlalchemy_test_model("test1")
    obj2 = sqlalchemy_test_model("test2")
    sqlalchemy_test_repository.add(obj1)
    sqlalchemy_test_repository.add(obj2)

    res = list(sqlalchemy_test_repository.find(EqualsSpecification("id", "no_result")))

    assert len(res) == 0


def test_find_with_spec_multiple_results(
    sqlalchemy_test_repository, sqlalchemy_test_model
):
    from fractal_specifications.generic.operators import EqualsSpecification

    obj1 = sqlalchemy_test_model("test1")
    obj2 = sqlalchemy_test_model("test2")
    sqlalchemy_test_repository.add(obj1)
    sqlalchemy_test_repository.add(obj2)

    res = list(sqlalchemy_test_repository.find(EqualsSpecification("name", "test")))

    assert len(res) == 2
    assert res == [obj1, obj2]


def test_find_with_and_spec(sqlalchemy_test_repository, sqlalchemy_test_model):
    from fractal_specifications.generic.operators import EqualsSpecification

    obj1 = sqlalchemy_test_model("test1")
    obj2 = sqlalchemy_test_model("test2")
    sqlalchemy_test_repository.add(obj1)
    sqlalchemy_test_repository.add(obj2)

    from fractal_specifications.generic.collections import AndSpecification

    res = list(
        sqlalchemy_test_repository.find(
            AndSpecification(
                [
                    EqualsSpecification("name", "test"),
                    EqualsSpecification("description", "test"),
                ]
            )
        )
    )

    assert len(res) == 2
    assert res == [obj1, obj2]


def test_find_with_or_spec(sqlalchemy_test_repository, sqlalchemy_test_model):
    from fractal_specifications.generic.collections import OrSpecification
    from fractal_specifications.generic.operators import EqualsSpecification

    obj1 = sqlalchemy_test_model("test1")
    obj2 = sqlalchemy_test_model("test2")
    sqlalchemy_test_repository.add(obj1)
    sqlalchemy_test_repository.add(obj2)

    res = list(
        sqlalchemy_test_repository.find(
            OrSpecification(
                [EqualsSpecification("id", "test1"), EqualsSpecification("id", "test2")]
            )
        )
    )

    assert len(res) == 2
    assert res == [obj1, obj2]


def test_update(sqlalchemy_test_repository, sqlalchemy_test_model):
    from fractal_specifications.generic.operators import EqualsSpecification

    obj = sqlalchemy_test_model("test")
    sqlalchemy_test_repository.add(obj)

    obj.name = "update"
    res = sqlalchemy_test_repository.update(obj)

    assert res == obj
    assert res.name == "update"

    res = sqlalchemy_test_repository.find_one(EqualsSpecification("id", "test"))

    assert res == obj
    assert res.name == "update"


def test_update_add_item(
    sqlalchemy_test_repository, sqlalchemy_test_model, sqlalchemy_test_sub_model
):
    from fractal_specifications.generic.operators import EqualsSpecification

    obj = sqlalchemy_test_model("test")
    sqlalchemy_test_repository.add(obj)

    sub_obj = sqlalchemy_test_sub_model("item", item_id=obj.id)
    obj.items = [sub_obj]
    res = sqlalchemy_test_repository.update(obj)

    assert res == obj
    assert res.items == [sub_obj]

    res = sqlalchemy_test_repository.find_one(EqualsSpecification("id", "test"))

    assert res == obj
    assert res.items == [sub_obj]


def test_update_delete_item(
    sqlalchemy_test_repository, sqlalchemy_test_model, sqlalchemy_test_sub_model
):
    from fractal_specifications.generic.operators import EqualsSpecification

    obj = sqlalchemy_test_model("test")
    sqlalchemy_test_repository.add(obj)
    sub_obj = sqlalchemy_test_sub_model("item", item_id=obj.id)
    obj.items = [sub_obj]
    sqlalchemy_test_repository.update(obj)

    obj.items = []
    res = sqlalchemy_test_repository.update(obj)

    assert res == obj
    assert res.items == []

    res = sqlalchemy_test_repository.find_one(EqualsSpecification("id", "test"))

    assert res == obj
    assert res.items == []


def test_update_item(
    sqlalchemy_test_repository, sqlalchemy_test_model, sqlalchemy_test_sub_model
):
    from fractal_specifications.generic.operators import EqualsSpecification

    obj = sqlalchemy_test_model("test")
    sqlalchemy_test_repository.add(obj)
    sub_obj = sqlalchemy_test_sub_model("item", item_id=obj.id)
    obj.items = [sub_obj]
    sqlalchemy_test_repository.update(obj)

    obj.items[0].name = "update"
    sqlalchemy_test_repository.update(obj)

    res = sqlalchemy_test_repository.find_one(EqualsSpecification("id", "test"))

    assert res == obj
    assert res.items[0].name == "update"


# TODO fix test for py39
# def test_update_item_error(
#     sqlalchemy_test_repository_error, sqlalchemy_test_model, sqlalchemy_test_sub_model
# ):
#     from fractal_specifications.contrib.sqlalchemy.repositories import UnknownListItemTypeException
#
#     obj = sqlalchemy_test_model("test")
#     sqlalchemy_test_repository_error.add(obj)
#     sub_obj = sqlalchemy_test_sub_model("item", item_id=obj.id)
#     obj.items = [sub_obj]
#
#     with pytest.raises(UnknownListItemTypeException):
#         sqlalchemy_test_repository_error.update(obj)


def test_update_upsert(sqlalchemy_test_repository, sqlalchemy_test_model):
    from fractal_specifications.generic.operators import EqualsSpecification

    obj = sqlalchemy_test_model("test")

    res = sqlalchemy_test_repository.update(obj, upsert=True)

    assert res == obj

    res = sqlalchemy_test_repository.find_one(EqualsSpecification("id", "test"))

    assert res == obj
    assert res.name == "test"


def test_update_upsert_error(sqlalchemy_test_repository, sqlalchemy_test_model):
    from fractal_specifications.generic.operators import EqualsSpecification

    obj = sqlalchemy_test_model("test")

    res = sqlalchemy_test_repository.update(obj, upsert=False)

    assert res is None

    res = sqlalchemy_test_repository.find_one(EqualsSpecification("id", "test"))

    assert res is None


def test_remove_one(sqlalchemy_test_repository, sqlalchemy_test_model):
    from fractal_specifications.generic.operators import EqualsSpecification

    obj1 = sqlalchemy_test_model("test1")
    obj2 = sqlalchemy_test_model("test2")
    sqlalchemy_test_repository.add(obj1)
    sqlalchemy_test_repository.add(obj2)

    sqlalchemy_test_repository.remove_one(EqualsSpecification("id", "test1"))

    res = list(sqlalchemy_test_repository.find())

    assert len(res) == 1
    assert res == [obj2]


def test_find_one(sqlalchemy_test_repository, sqlalchemy_test_model):
    from fractal_specifications.generic.operators import EqualsSpecification

    obj1 = sqlalchemy_test_model("test1")
    obj2 = sqlalchemy_test_model("test2")
    sqlalchemy_test_repository.add(obj1)
    sqlalchemy_test_repository.add(obj2)

    res = sqlalchemy_test_repository.find_one(EqualsSpecification("id", "test2"))

    assert res == obj2



from tests.fixtures.contrib.django import *  # NOQA NOSONAR
from tests.fixtures.contrib.fastapi import *  # NOQA NOSONAR
from tests.fixtures.contrib.gcp import *  # NOQA NOSONAR
from tests.fixtures.contrib.mongo import *  # NOQA NOSONAR
from tests.fixtures.contrib.roles import *  # NOQA NOSONAR
from tests.fixtures.contrib.sqlalchemy import *  # NOQA NOSONAR
from tests.fixtures.contrib.tokens import *  # NOQA NOSONAR



import json
import os
import uuid
from typing import Iterator, Optional

from fractal_specifications.generic.operators import (
    EqualsSpecification,
    NotSpecification,
)
from fractal_specifications.generic.specification import Specification

from fractal.core.exceptions import ObjectNotFoundException
from fractal.core.repositories import Entity, FileRepository, Repository
from fractal.core.utils.json_encoder import EnhancedEncoder


class RootDirMixin(object):
    def __init__(self, root_dir: str, *args, **kwargs):
        super(RootDirMixin, self).__init__(*args, **kwargs)

        self.root_dir = root_dir


class FileRepositoryMixin(RootDirMixin, Repository[Entity]):
    @property
    def _filename(self) -> str:
        return os.path.join(self.root_dir, "db", f"{self.__class__.__name__}.txt")

    @property
    def _entities(self):
        if not os.path.exists(self._filename):
            return None
        with open(self._filename, "r") as fp:
            for line in fp.readlines():
                yield self.entity(**json.loads(line))

    def add(self, entity: Entity) -> Entity:
        with open(self._filename, "a") as fp:
            fp.write(json.dumps(entity.asdict(), cls=EnhancedEncoder) + "\n")
        return entity

    def update(self, entity: Entity, upsert=False) -> Entity:
        try:
            current = self.find_one(EqualsSpecification("id", entity.id))
        except ObjectNotFoundException:
            current = None
        if current or upsert:
            self.remove_one(EqualsSpecification("id", entity.id))
            return self.add(entity)

    def remove_one(self, specification: Specification):
        entities = list(self.find(NotSpecification(specification)))
        with open(self._filename, "w") as fp:
            fp.writelines(
                [json.dumps(e.asdict(), cls=EnhancedEncoder) + "\n" for e in entities]
            )

    def find_one(self, specification: Specification) -> Optional[Entity]:
        for entity in filter(
            lambda i: specification.is_satisfied_by(i), self._entities
        ):
            return entity
        if self.object_not_found_exception:
            raise self.object_not_found_exception
        raise ObjectNotFoundException(f"{self.entity.__name__} not found!")

    def find(self, specification: Optional[Specification] = None) -> Iterator[Entity]:
        if specification:
            entities = filter(
                lambda i: specification.is_satisfied_by(i), self._entities
            )
        else:
            entities = self._entities
        for entity in entities:
            yield entity

    def is_healthy(self) -> bool:
        return True


class FileFileRepositoryMixin(RootDirMixin, FileRepository):
    def upload_file(self, data: bytes, content_type: str, reference: str = "") -> str:
        if not reference:
            reference = str(uuid.uuid4())
        with open(os.path.join(self.root_dir, "media", reference), "wb") as fp:
            fp.write(data)
        return reference

    def get_file(self, reference: str) -> bytes:
        with open(os.path.join(self.root_dir, "media", reference), "rb") as fp:
            return fp.read()

    def delete_file(self, reference: str) -> bool:
        try:
            os.remove(os.path.join(self.root_dir, "media", reference))
            return True
        except Exception:
            return False



import pytest


def pytest_configure(config):
    config.addinivalue_line("markers", "context: overrides for adapters in Context")


def reset(fractal, settings):
    fractal.settings.reload(settings)
    fractal.context.reload()
    return fractal


@pytest.fixture(autouse=True)
def fractal(request, empty_application_context, settings):
    from fractal.core.utils.fractal import ApplicationFractal

    fractal = ApplicationFractal(empty_application_context, settings)

    marker = request.node.get_closest_marker("settings")
    if marker:
        return reset(fractal, marker.args[0])
    else:
        return reset(fractal, {})


@pytest.fixture(autouse=True)
def token_roles_service_fractal(request, token_service_application_context, settings):
    from fractal.core.utils.fractal import ApplicationFractal

    fractal = ApplicationFractal(token_service_application_context, settings)

    marker = request.node.get_closest_marker("settings")
    if marker:
        return reset(fractal, marker.args[0])
    else:
        return reset(fractal, {})


@pytest.fixture
def reset_fractal(token_roles_service_fractal):
    def _reset(settings):
        return reset(token_roles_service_fractal, settings)

    return _reset



from abc import ABC

from fractal import ApplicationContext


class Service(ABC):
    @classmethod
    def install(cls, context: ApplicationContext):
        yield cls()

    def is_healthy(self) -> bool:
        return True



from dataclasses import dataclass
from typing import List, Optional

from fastapi import Depends
from fastapi.security import OAuth2PasswordBearer
from fractal_roles.models import Role, TokenPayloadRolesMixin
from fractal_roles.services import RolesService as BaseRolesService
from fractal_tokens.services.dummy import DummyJsonTokenService
from fractal_tokens.services.generic import TokenService
from fractal_tokens.services.jwt import TokenPayload as BaseTokenPayload
from fractal_tokens.services.jwt.automatic import AutomaticJwtTokenService

from fractal import ApplicationContext, Fractal
from fractal.core.models import Model
from fractal.core.services import Service

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="login")


@dataclass
class TokenPayload(Model, BaseTokenPayload):
    ...


@dataclass
class TokenPayloadRoles(Model, TokenPayloadRolesMixin, BaseTokenPayload):
    roles: Optional[List[Role]] = None
    account: str = ""

    def __post_init__(self):
        if not self.roles:
            self.roles = []


class FractalAutomaticJwtTokenService(AutomaticJwtTokenService, Service):
    token_payload_cls = TokenPayloadRoles

    @classmethod
    def install(cls, context: ApplicationContext):
        yield cls(
            issuer=context.settings.JWK_SERVICE_HOST,
            secret_key=context.settings.SECRET_KEY,
            jwk_service=context.jwk_service,
        )


class FractalDummyJsonTokenService(DummyJsonTokenService, Service):
    token_payload_cls = TokenPayloadRoles

    @classmethod
    def install(cls, context: ApplicationContext):
        yield cls()


def get_payload(fractal: Fractal, *, typ: str = "access"):
    assert hasattr(fractal, "context")
    context = getattr(fractal, "context")
    assert hasattr(context, "token_service")

    def _get_payload(token: str = Depends(oauth2_scheme)) -> TokenPayload:
        return context.token_service.verify(token, typ=typ)

    return _get_payload


def get_payload_roles(
    fractal: Fractal, *, endpoint: str = "", method: str = "get", typ: str = "access"
):
    assert hasattr(fractal, "context")
    context = getattr(fractal, "context")
    assert hasattr(context, "token_service")
    assert hasattr(context, "roles_service")

    def _get_payload(token: str = Depends(oauth2_scheme)) -> TokenPayloadRoles:
        payload = context.token_service.verify(token, typ=typ)
        payload = context.roles_service.verify(payload, endpoint, method)
        return payload

    return _get_payload


def get_token_roles_payload(
    *,
    token_service: TokenService,
    roles_service: BaseRolesService,
    endpoint: str = "",
    method: str = "get",
    typ: str = "access",
):
    def _get_payload(token: str = Depends(oauth2_scheme)) -> TokenPayloadRoles:
        payload = token_service.verify(token, typ=typ)
        payload = roles_service.verify(payload, endpoint, method)
        return payload

    return _get_payload



from dataclasses import asdict
from typing import Dict, Generator, Optional, Type

from django.db.models import ForeignKey, Model, Q
from fractal_specifications.contrib.django.specifications import (
    DjangoOrmSpecificationBuilder,
)
from fractal_specifications.generic.specification import Specification

from fractal.core.repositories import Entity, Repository


class DjangoModelRepositoryMixin(Repository[Entity]):
    def __init__(self, django_model: Type[Model], domain_model: Entity):
        self.django_model = django_model
        self.domain_model = domain_model

    def add(self, entity: Entity) -> Entity:
        direct_data, related_data = self.__get_direct_related_data(entity)
        obj = self.django_model.objects.create(**direct_data)
        entity.id = obj.id
        return entity

    def update(self, entity: Entity, upsert=False) -> Entity:
        if entities := self.django_model.objects.filter(pk=entity.id):
            direct_data, related_data = self.__get_direct_related_data(entity)
            entities.update(**direct_data)
            return self.django_model.objects.get(pk=entity.id)
        elif upsert:
            return self.add(entity)

    def __get_direct_related_data(self, entity: Entity):
        direct_data = {}
        related_data = {}

        def field_name(field):
            if type(field) is ForeignKey:
                return field.name + "_id"
            return field.name

        direct_fields = [field_name(f) for f in self.django_model._meta.fields]
        for k, v in asdict(entity).items():
            if type(v) == list:
                related_data[k] = v
            else:
                if k in direct_fields:
                    direct_data[k] = v
        return direct_data, related_data

    def __get_obj(self, specification: Specification):
        _filter = DjangoOrmSpecificationBuilder.build(specification)
        if type(_filter) is list:
            obj = self.django_model.objects.get(*_filter)
        elif type(_filter) is dict:
            obj = self.django_model.objects.get(**_filter)
        elif type(_filter) is Q:
            obj = self.django_model.objects.get(_filter)
        else:
            raise self.django_model.DoesNotExist
        return obj

    def remove_one(self, specification: Specification):
        self.__get_obj(specification).delete()

    def find_one(self, specification: Specification) -> Optional[Entity]:
        return self._obj_to_domain(self.__get_obj(specification).__dict__)

    def find(
        self, specification: Specification = None
    ) -> Generator[Entity, None, None]:
        _filter = DjangoOrmSpecificationBuilder.build(specification)
        if type(_filter) is list:
            queryset = self.django_model.objects.filter(*_filter)
        elif type(_filter) is dict:
            queryset = self.django_model.objects.filter(**_filter)
        elif type(_filter) is Q:
            queryset = self.django_model.objects.filter(_filter)
        else:
            queryset = self.django_model.objects.all()

        for obj in queryset:
            yield self._obj_to_domain(obj.__dict__)

    def is_healthy(self) -> bool:
        return True

    def _obj_to_domain(self, obj: Dict) -> Entity:
        return self.domain_model.clean(**obj)



from typing import List, Optional

from fractal.core.process.action import Action
from fractal.core.process.process_scope import ProcessScope


class Process:
    def __init__(self, actions: List[Action]):
        self.actions = actions

    def run(self, scope: Optional[ProcessScope] = None) -> ProcessScope:
        if not scope:
            scope = ProcessScope()
        for action in self.actions:
            scope.update(action.execute(scope))
        return scope



from pydantic import BaseModel

from fractal.core.utils.json_encoder import EnhancedEncoder


class BaseModelEnhancedEncoder(EnhancedEncoder):
    def default(self, o):
        if isinstance(o, BaseModel):
            return o.dict()
        return super(BaseModelEnhancedEncoder, self).default(o)



def test_load_data_dict(external_data_inmemory_repository, an_object):
    data = {an_object.__class__.__name__.lower(): [an_object.__dict__]}

    external_data_inmemory_repository.load_data_dict(data)

    assert list(external_data_inmemory_repository.find()) == [an_object]


def test_dump_data_dict(external_data_inmemory_repository, an_object):
    external_data_inmemory_repository.add(an_object)

    data = (an_object.__class__.__name__.lower(), [an_object.__dict__])

    assert external_data_inmemory_repository.dump_data_dict() == data


def test_load_data_json(external_data_inmemory_repository, an_object):
    import json

    data = {an_object.__class__.__name__.lower(): json.dumps([an_object.__dict__])}

    external_data_inmemory_repository.load_data_json(data)

    assert list(external_data_inmemory_repository.find()) == [an_object]


def test_dump_data_json(external_data_inmemory_repository, an_object):
    external_data_inmemory_repository.add(an_object)

    import json

    data = (an_object.__class__.__name__.lower(), json.dumps([an_object.__dict__]))

    assert external_data_inmemory_repository.dump_data_json() == data



from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Type
from uuid import UUID

from fastapi import APIRouter, Depends, status
from fractal_specifications.generic.specification import Specification
from fractal_tokens.services.generic import TokenPayload
from pydantic import BaseModel

from fractal import Fractal
from fractal.contrib.fastapi.routers import Routes
from fractal.contrib.fastapi.routers.domain.models import AdapterInfo, Info
from fractal.contrib.fastapi.routers.tokens import (
    TokenPayloadRoles,
    get_payload,
    get_payload_roles,
)
from fractal.core.models import Model
from fractal.core.services import Service


def inject_default_routes(fractal: Fractal):
    router = APIRouter()

    @router.get(Routes.ROOT)
    def read_root():
        return {
            "FastAPI": fractal.settings.APP_NAME
            if fractal.settings and hasattr(fractal.settings, "APP_NAME")
            else "Fractal Service",
        }

    @router.get(Routes.INFO, responses={200: {"model": Info}})
    def info(
        payload: TokenPayload = Depends(get_payload(fractal)),
    ):
        adapters = list(fractal.context.adapters())
        data = [
            AdapterInfo(
                adapter=adapter.__class__.__name__,
                status=False,
            )
            for adapter in adapters
        ]
        for i in range(len(data)):
            try:
                data[i].status_ok = adapters[i].is_healthy()
            except Exception as e:
                fractal.context.logger.error(e)
                data[i].status_ok = False
        return data

    return router


class DefaultRestRouterService(Service, ABC):
    domain_entity_class: Type[Model]
    entities_route: str
    entity_route: str
    entities_endpoint: str
    entity_endpoint: str
    entity_contract: Type[BaseModel]
    create_entity_contract: Type[BaseModel]

    @abstractmethod
    def add_entity(
        self,
        contract: BaseModel,
        **kwargs,
    ):
        raise NotImplementedError

    @abstractmethod
    def find_entities(
        self,
        q: str = "",
        **kwargs,
    ):
        raise NotImplementedError

    @abstractmethod
    def get_entity(
        self,
        entity_id: UUID,
        **kwargs,
    ):
        raise NotImplementedError

    @abstractmethod
    def update_entity(
        self,
        entity_id: UUID,
        contract: BaseModel,
        **kwargs,
    ):
        raise NotImplementedError

    @abstractmethod
    def delete_entity(
        self,
        entity_id: UUID,
        **kwargs,
    ) -> Dict:
        raise NotImplementedError


class Contract(BaseModel, ABC):
    @abstractmethod
    def to_entity(self, **kwargs):
        raise NotImplementedError


class BasicRestRouterService(DefaultRestRouterService):
    def add_entity(
        self,
        contract: Contract,
        *,
        specification: Specification = None,
        **kwargs,
    ):
        try:
            UUID(contract.id)
        except (ValueError, TypeError):
            contract.id = None
        _entity = contract.to_entity(
            user_id=kwargs.get("sub"), account_id=kwargs.get("account")
        )
        return self.ingress_service.add(
            _entity, str(kwargs.get("sub")), specification=specification
        )

    def find_entities(
        self,
        q: str = "",
        *,
        specification: Specification = None,
        **kwargs,
    ):
        return self.ingress_service.find(
            str(kwargs.get("account")), q, specification=specification
        )

    def get_entity(
        self,
        entity_id: UUID,
        *,
        specification: Specification = None,
        **kwargs,
    ):
        return self.ingress_service.get(
            str(entity_id), str(kwargs.get("account")), specification=specification
        )

    def update_entity(
        self,
        entity_id: UUID,
        contract: Contract,
        *,
        specification: Specification = None,
        **kwargs,
    ):
        if _entity := self.get_entity(entity_id, specification=specification, **kwargs):
            _entity = _entity.update(
                {k: v for k, v in contract.dict().items() if v is not None}
            )
        else:
            _entity = contract.to_entity(
                id=entity_id,
                user_id=kwargs.get("sub"),
                account_id=kwargs.get("account"),
            )
        return self.ingress_service.update(
            str(entity_id), _entity, str(kwargs.get("sub")), specification=specification
        )

    def delete_entity(
        self,
        entity_id: UUID,
        *,
        specification: Specification = None,
        **kwargs,
    ) -> Dict:
        self.ingress_service.delete(
            str(entity_id),
            str(kwargs.get("sub")),
            str(kwargs.get("account")),
            specification=specification,
        )
        return {}


def inject_default_rest_routes(
    fractal: Fractal,
    *,
    router_service_class: Type[DefaultRestRouterService],
):
    router = APIRouter()

    @router.post(
        router_service_class().entities_route,
        response_model=router_service_class().domain_entity_class,
        status_code=status.HTTP_201_CREATED,
        name=f"Add {router_service_class().domain_entity_class.__name__}",
    )
    def add_entity(
        entity: router_service_class().create_entity_contract,
        payload: TokenPayloadRoles = Depends(
            get_payload_roles(
                fractal,
                endpoint=router_service_class().entities_endpoint,
                method="post",
            )
        ),
    ):
        return router_service_class().add_entity(
            contract=entity,
            specification=payload.specification,
            **payload.dict(),
        )

    @router.get(
        router_service_class().entities_route,
        response_model=List[router_service_class().domain_entity_class],
        status_code=status.HTTP_200_OK,
        name=f"Find {router_service_class().domain_entity_class.__name__} entities",
    )
    def find_entities(
        q: Optional[str] = "",
        payload: TokenPayloadRoles = Depends(
            get_payload_roles(
                fractal,
                endpoint=router_service_class().entities_endpoint,
                method="get",
            )
        ),
    ):
        return router_service_class().find_entities(
            q=q,
            specification=payload.specification,
            **payload.dict(),
        )

    @router.get(
        router_service_class().entity_route,
        response_model=router_service_class().domain_entity_class,
        status_code=status.HTTP_200_OK,
        name=f"Get {router_service_class().domain_entity_class.__name__} entity",
    )
    def get_entity(
        entity_id: UUID,
        payload: TokenPayloadRoles = Depends(
            get_payload_roles(
                fractal,
                endpoint=router_service_class().entities_endpoint,
                method="get",
            )
        ),
    ):
        return router_service_class().get_entity(
            entity_id=entity_id,
            specification=payload.specification,
            **payload.dict(),
        )

    @router.put(
        router_service_class().entity_route,
        response_model=router_service_class().domain_entity_class,
        status_code=status.HTTP_200_OK,
        name=f"Update {router_service_class().domain_entity_class.__name__}",
    )
    def update_entity(
        entity_id: UUID,
        entity: router_service_class().entity_contract,
        payload: TokenPayloadRoles = Depends(
            get_payload_roles(
                fractal,
                endpoint=router_service_class().entities_endpoint,
                method="put",
            )
        ),
    ):
        return router_service_class().update_entity(
            entity_id=entity_id,
            contract=entity,
            specification=payload.specification,
            **payload.dict(),
        )

    @router.delete(
        router_service_class().entity_route,
        response_model=Dict,
        status_code=status.HTTP_200_OK,
        name=f"Delete {router_service_class().domain_entity_class.__name__}",
    )
    def delete_entity(
        entity_id: UUID,
        payload: TokenPayloadRoles = Depends(
            get_payload_roles(
                fractal,
                endpoint=router_service_class().entities_endpoint,
                method="delete",
            )
        ),
    ) -> Dict:
        return router_service_class().delete_entity(
            entity_id=entity_id,
            specification=payload.specification,
            **payload.dict(),
        )

    return router



from app.context import ApplicationContext
from app.settings import Settings

from fractal import Fractal


class ApplicationFractal(Fractal):
    settings = Settings()
    context = ApplicationContext()



import uuid

from fractal.core.event_sourcing.event import SendingEvent
from fractal.core.event_sourcing.event_projector import EventProjector
from fractal.core.event_sourcing.event_store import EventStore
from fractal.core.event_sourcing.event_stream import EventStream


class EventStoreProjector(EventProjector):
    def __init__(self, event_store: EventStore):
        self.event_store = event_store

    def project(self, id: str, event: SendingEvent):
        self.event_store.commit(
            EventStream(id=str(uuid.uuid4()), events=[event]), aggregate="", version=1
        )



import uuid
from typing import Dict, Iterator, Optional

from fractal_specifications.generic.specification import Specification

from fractal.core.exceptions import ObjectNotFoundException
from fractal.core.repositories import Entity, FileRepository, Repository


class InMemoryRepositoryMixin(Repository[Entity]):
    def __init__(self):
        super(InMemoryRepositoryMixin, self).__init__()

        self.entities: Dict[str, Entity] = {}

    def add(self, entity: Entity) -> Entity:
        self.entities[entity.id] = entity
        return entity

    def update(self, entity: Entity, upsert=False) -> Entity:
        if entity.id in self.entities or upsert:
            return self.add(entity)

    def remove_one(self, specification: Specification):
        if obj := self.find_one(specification):
            del self.entities[obj.id]

    def find_one(self, specification: Specification) -> Optional[Entity]:
        for entity in filter(
            lambda i: specification.is_satisfied_by(i), self.entities.values()
        ):
            return entity
        if self.object_not_found_exception:
            raise self.object_not_found_exception
        raise ObjectNotFoundException(f"{self.entity.__name__} not found!")

    def find(self, specification: Optional[Specification] = None) -> Iterator[Entity]:
        if specification:
            entities = filter(
                lambda i: specification.is_satisfied_by(i), self.entities.values()
            )
        else:
            entities = self.entities.values()
        for entity in entities:
            yield entity

    def is_healthy(self) -> bool:
        return True


class InMemoryFileRepositoryMixin(FileRepository[Entity]):
    def __init__(self):
        super(InMemoryFileRepositoryMixin, self).__init__()

        self.files: Dict[str, bytes] = {}

    def upload_file(self, data: bytes, content_type: str, reference: str = "") -> str:
        if not reference:
            reference = str(uuid.uuid4())
        self.files[reference] = data
        return reference

    def get_file(self, reference: str) -> bytes:
        return self.files.get(reference, "")

    def delete_file(self, reference: str) -> bool:
        if reference in self.files:
            del self.files[reference]
            return True
        return False



import os
from datetime import datetime

import pytest


@pytest.fixture
def settings_class():
    from fractal.core.utils.settings import Settings

    class FakeSettings(Settings):
        APP_NAME = os.getenv("APP_NAME", "app_name")
        BASE_DIR = os.getcwd()
        ROOT_DIR = os.getcwd()

        def load(self):
            self.WEBSITE_HOST = os.getenv("WEBSITE_HOST", "http://localhost:8000")

    return FakeSettings


@pytest.fixture
def settings(settings_class):
    return settings_class()


@pytest.fixture
def fake_application_context_class(
    inmemory_repository, fake_service_class, another_fake_service_class, settings_class
):
    from fractal.core.utils.application_context import ApplicationContext

    class FakeApplicationContext(ApplicationContext):
        settings = settings_class()

        def load_internal_services(self):
            super(FakeApplicationContext, self).load_internal_services()

            if os.getenv("FAKE_SERVICE", "") == "another":
                self.fake_service = self.install_service(
                    another_fake_service_class, name="fake_service"
                )()
            else:
                self.fake_service = self.install_service(fake_service_class)()

        def load_repositories(self):
            super(FakeApplicationContext, self).load_repositories()

            self.install_repository(inmemory_repository)

    return FakeApplicationContext


@pytest.fixture
def empty_application_context(settings_class):
    from fractal.core.utils.application_context import ApplicationContext

    class EmptyApplicationContext(ApplicationContext):
        settings = settings_class()

    return EmptyApplicationContext()


@pytest.fixture
def now():
    return datetime.utcnow()



from __future__ import annotations

import dataclasses
import logging
from abc import ABC, abstractmethod
from typing import Dict, Generator, Generic, List, Optional, TypeVar, get_type_hints

from fractal_specifications.contrib.sqlalchemy.specifications import (
    SqlAlchemyOrmSpecificationBuilder,
)
from fractal_specifications.generic.operators import EqualsSpecification
from fractal_specifications.generic.specification import Specification
from sqlalchemy import MetaData, Table, create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.exc import ArgumentError, IntegrityError
from sqlalchemy.orm import Mapper, Session, sessionmaker  # NOQA

from fractal.core.exceptions import DomainException
from fractal.core.repositories import Entity, Repository

EntityDao = TypeVar("EntityDao")


class UnknownListItemTypeException(DomainException):
    code = "UNKNOWN_LIST_ITEM_TYPE_EXCEPTION"
    status_code = 500


class IntegrityErrorException(DomainException):
    code = "INTEGRITY_ERROR_EXCEPTION"
    status_code = 500


class SqlAlchemyDao(ABC):
    @staticmethod
    @abstractmethod
    def mapper(meta: MetaData, foreign_keys: Dict[str, Mapper]) -> Mapper:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def table(meta: MetaData) -> Table:
        raise NotImplementedError

    @staticmethod
    def from_domain(obj: Entity) -> SqlAlchemyDao:
        raise NotImplementedError


class DaoMapper(ABC):
    instance = None

    def __new__(cls, *args, **kwargs):
        if not isinstance(cls.instance, cls):
            cls.instance = object.__new__(cls)
            cls.instance.done = False
        return cls.instance

    def start_mappers(self, engine: Engine):
        if not self.done:
            meta = MetaData()
            self.application_mappers(meta)
            meta.create_all(engine)
            self.done = True

    @abstractmethod
    def application_mappers(self, meta: MetaData):
        raise NotImplementedError


class AbstractUnitOfWork(ABC):
    def __enter__(self) -> AbstractUnitOfWork:
        return self

    def __exit__(self, *args):
        self.rollback()

    @abstractmethod
    def commit(self):
        raise NotImplementedError

    @abstractmethod
    def rollback(self):
        raise NotImplementedError


class SqlAlchemyUnitOfWork(AbstractUnitOfWork):
    def __init__(self):
        self.session_factory = None

    def __enter__(self) -> AbstractUnitOfWork:
        self.session = self.session_factory()  # type: Session
        return super().__enter__()

    def __exit__(self, *args):
        super().__exit__(*args)
        self.session.close()

    def commit(self):
        self.session.commit()

    def rollback(self):
        self.session.rollback()


class SqlAlchemyRepositoryMixin(
    Generic[Entity, EntityDao], Repository[Entity], SqlAlchemyUnitOfWork
):
    entity = Entity
    entity_dao = EntityDao
    application_mapper = DaoMapper

    def __init__(self, connection_str: str):
        super().__init__()

        self.connection_str = connection_str
        engine = create_engine(
            self.connection_str,
        )

        self.application_mapper().start_mappers(engine)

        self.session_factory = sessionmaker(
            bind=engine,
            expire_on_commit=False,
        )

    def add(self, entity: Entity) -> Entity:
        return self.__add(entity, self.entity_dao)

    def __add(self, entity: Entity, entity_dao_class: SqlAlchemyDao) -> Entity:
        entity_dao = entity_dao_class.from_domain(entity)
        with self:
            try:
                self.session.add(entity_dao)
                self.commit()
            except IntegrityError as e:
                raise IntegrityErrorException(e)
        return entity

    def update(self, entity: Entity, *, upsert=False) -> Entity:
        return self.__update(entity, self.entity_dao, upsert=upsert)

    def __update(
        self, entity: Entity, entity_dao_class: SqlAlchemyDao, *, upsert=False
    ) -> Entity:
        """Recursive function"""
        with self:
            try:
                existing_entity_dao = self._find_one_raw(
                    EqualsSpecification("id", entity.id),
                    entity_dao_class=entity_dao_class,
                )
            except ArgumentError:
                raise UnknownListItemTypeException(
                    f"DAO '{entity_dao_class}' has an unknown list collection DAO, please add the type for the list."
                )
            if existing_entity_dao:
                self.__update_existing_record(
                    entity, entity_dao_class, existing_entity_dao
                )
                return entity
            elif upsert:
                return self.__add(entity, entity_dao_class)

    def __update_existing_record(self, entity, entity_dao_class, existing_entity_dao):
        updating_entity_dao = entity_dao_class.from_domain(entity)
        regular_fields = []
        list_fields = []

        for k, v in entity_dao_class.__annotations__.items():
            if hasattr(v, "__origin__") and v.__origin__ is list:
                list_fields.append(k)
            else:
                regular_fields.append(k)

        self.__update_main_entity(
            existing_entity_dao, regular_fields, updating_entity_dao
        )
        self.__update_compound_entities(entity, entity_dao_class, list_fields)
        self.commit()

    def __update_main_entity(
        self, existing_entity_dao, regular_fields, updating_entity_dao
    ):
        for k, v in updating_entity_dao.__dict__.items():
            if hasattr(existing_entity_dao, k) and k in regular_fields:
                setattr(existing_entity_dao, k, v)

    def __update_compound_entities(self, entity, entity_dao_class, list_fields):
        for field in list_fields:
            item_dao_class = get_type_hints(entity_dao_class)[field].__args__[0]

            # check for new items
            for item in getattr(entity, field):
                self.__update(item, item_dao_class, upsert=True)

            # check for items to delete
            foreign_key = getattr(entity_dao_class, field).expression.right
            items = list(
                self.session.query(item_dao_class).filter(foreign_key == entity.id)
            )
            item_ids = [item.id for item in getattr(entity, field)]
            for item_dao in items:
                if item_dao.id not in item_ids:
                    self.session.delete(item_dao)

    def remove_one(self, specification: Specification):
        entity = self._find_one_raw(specification)
        if entity:
            self.session.delete(entity)
            self.commit()

    def find_one(self, specification: Specification) -> Optional[Entity]:
        entity = self._find_one_raw(specification)
        if entity:
            return self._dao_to_domain(entity)

    def find(
        self, specification: Optional[Specification] = None
    ) -> Generator[Entity, None, None]:
        entities = self._find_raw(specification)

        if specification:
            entities = filter(lambda i: specification.is_satisfied_by(i), entities)
        for entity in entities:
            yield self._dao_to_domain(entity)

    def _dao_to_domain(self, entity: Entity):
        return self.__dao_to_domain(entity, self.entity, self.entity_dao)

    def __dao_to_domain(
        self, entity: Entity, domain_model: Entity, entity_dao: SqlAlchemyDao
    ):
        """Recursive function"""
        list_fields = []
        for k, v in entity_dao.__annotations__.items():
            if hasattr(v, "__origin__") and v.__origin__ is list:
                list_fields.append(k)
        d = entity.__dict__
        for field in list_fields:
            if hasattr(entity, field):
                item_domain_model = get_type_hints(domain_model)[field].__args__[0]
                item_entity_dao = get_type_hints(entity_dao)[field].__args__[0]
                d[field] = [
                    self.__dao_to_domain(sub_entity, item_domain_model, item_entity_dao)
                    for sub_entity in getattr(entity, field)
                ]
        fields = set(f.name for f in dataclasses.fields(domain_model))
        return domain_model(**{k: v for k, v in d.items() if k in fields})

    def _find_one_raw(
        self,
        specification: Specification,
        *,
        entity_dao_class: Optional[SqlAlchemyDao] = None,
    ) -> Optional[Entity]:
        entities = self._find_raw(specification, entity_dao_class=entity_dao_class)

        for entity in filter(lambda i: specification.is_satisfied_by(i), entities):
            return entity

    def _find_raw(
        self,
        specification: Optional[Specification],
        *,
        entity_dao_class: Optional[SqlAlchemyDao] = None,
    ) -> List[Entity]:
        _filter = {}
        if specification:
            _filter = SqlAlchemyOrmSpecificationBuilder.build(specification)
        if isinstance(_filter, list):
            entities = []
            for f in _filter:
                entities.extend(
                    list(
                        self.session.query(
                            entity_dao_class or self.entity_dao
                        ).filter_by(**dict(f))
                    )
                )
            return entities
        else:
            return self.session.query(entity_dao_class or self.entity_dao).filter_by(
                **dict(_filter)
            )

    def is_healthy(self) -> bool:
        try:
            with self:
                self.session.execute("SELECT 1")
        except Exception as e:
            logging.exception(f"Database is unhealthy! {e}")
            return False
        return True



from typing import Optional

from fractal_specifications.generic.specification import Specification

from fractal.core.process.action import Action
from fractal.core.process.process_scope import ProcessScope


class SetValueAction(Action):
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def execute(self, scope: ProcessScope) -> ProcessScope:
        return scope.update(ProcessScope(self.kwargs))


class IncreaseValueAction(Action):
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def execute(self, scope: ProcessScope) -> ProcessScope:
        scope[self.name] += self.value
        return scope


class PrintAction(Action):
    def __init__(self, text):
        self.text = text

    def execute(self, scope: ProcessScope) -> ProcessScope:
        print(self.text)
        return scope


class PrintValueAction(Action):
    def __init__(self, name):
        self.name = name

    def execute(self, scope: ProcessScope) -> ProcessScope:
        print(scope[self.name])
        return scope


class AddEntityAction(Action):
    def __init__(self, repository_key: str = "repository", **entity_defaults):
        self.repository_key = repository_key
        self.entity_defaults = entity_defaults

    def execute(self, scope: ProcessScope) -> ProcessScope:
        entity_class = scope[self.repository_key].entity
        data = self.entity_defaults
        if hasattr(scope, "contract"):
            data.update(scope["contract"])
        entity = entity_class(**data)
        scope["entity"] = scope[self.repository_key].add(entity)
        return scope


class UpdateEntityAction(Action):
    def __init__(self, repository_key: str = "repository"):
        self.repository_key = repository_key

    def execute(self, scope: ProcessScope) -> ProcessScope:
        entity = scope["entity"].update(scope["contract"])
        scope["entity"] = scope[self.repository_key].update(entity)
        return scope


class FetchEntityAction(Action):
    def __init__(
        self, specification: Specification, repository_key: str = "repository"
    ):
        self.specification = specification
        self.repository_key = repository_key

    def execute(self, scope: ProcessScope) -> ProcessScope:
        scope["entity"] = scope[self.repository_key].find_one(self.specification)
        return scope


class FindEntitiesAction(Action):
    def __init__(
        self,
        specification: Optional[Specification] = None,
        repository_key: str = "repository",
    ):
        self.specification = specification
        self.repository_key = repository_key

    def execute(self, scope: ProcessScope) -> ProcessScope:
        scope["entities"] = scope["entity"] = scope[self.repository_key].find(
            self.specification
        )
        return scope


class DeleteEntityAction(Action):
    def __init__(
        self, specification: Specification, repository_key: str = "repository"
    ):
        self.specification = specification
        self.repository_key = repository_key

    def execute(self, scope: ProcessScope) -> ProcessScope:
        scope[self.repository_key].remove_one(self.specification)
        return scope



import pytest

strings = [
    ("", ""),
    ("A", "a"),
    ("AaAaAa", "aa_aa_aa"),
]


@pytest.mark.parametrize("camel, snake", strings + [("AAA", "aaa")])
def test_camel_to_snake(camel, snake):
    from fractal.core.utils.string import camel_to_snake

    assert camel_to_snake(camel) == snake


@pytest.mark.parametrize("camel, snake", strings + [("Aaa", "aaa")])
def test_snake_to_camel(camel, snake):
    from fractal.core.utils.string import snake_to_camel

    assert snake_to_camel(snake) == camel



from dataclasses import dataclass
from datetime import datetime
from typing import Any


@dataclass
class Message:
    id: str
    occurred_on: datetime
    event: str
    data: Any
    object_id: str
    aggregate_root_id: str



from abc import ABC
from dataclasses import dataclass, field
from typing import Dict, List

import pytest


@pytest.fixture
def sqlalchemy_test_model(sqlalchemy_test_sub_model):
    @dataclass
    class TestModel:
        id: str
        name: str = "test"
        description: str = "test"
        items: List[sqlalchemy_test_sub_model] = field(default_factory=list)

    return TestModel


@pytest.fixture
def sqlalchemy_test_sub_model():
    @dataclass
    class TestSubModel:
        id: str
        name: str = "test"
        item_id: str = ""

    return TestSubModel


@pytest.fixture
def sqlalchemy_test_model_dao(sqlalchemy_test_model, sqlalchemy_test_sub_model_dao):
    from sqlalchemy import Column, MetaData, String, Table
    from sqlalchemy.orm import Mapper, mapper

    from fractal.contrib.sqlalchemy.repositories import SqlAlchemyDao

    @dataclass
    class TestModelDao(SqlAlchemyDao):
        id: str
        name: str
        description: str
        items: List[sqlalchemy_test_sub_model_dao] = field(default_factory=list)

        @staticmethod
        def from_domain(obj: sqlalchemy_test_model):
            dao = TestModelDao(
                id=obj.id,
                name=obj.name,
                description=obj.description,
            )
            dao.items = [
                sqlalchemy_test_sub_model_dao.from_domain(e) for e in obj.items
            ]
            return dao

        @staticmethod
        def mapper(meta: MetaData, foreign_keys: Dict[str, Mapper]) -> Mapper:
            return mapper(
                TestModelDao,
                TestModelDao.table(meta),
            )

        @staticmethod
        def table(meta: MetaData) -> Table:
            return Table(
                "test_model",
                meta,
                Column("id", String(), primary_key=True),
                Column("name", String()),
                Column("description", String()),
            )

    return TestModelDao


@pytest.fixture
def sqlalchemy_test_model_dao_error(
    sqlalchemy_test_model, sqlalchemy_test_sub_model_dao
):
    from fractal_specifications.contrib.sqlalchemy.repositories import SqlAlchemyDao
    from sqlalchemy import Column, MetaData, String, Table
    from sqlalchemy.orm import Mapper, mapper

    @dataclass
    class TestModelDao(SqlAlchemyDao):
        id: str
        name: str
        description: str
        items: List = field(
            default_factory=list
        )  # no typing for list items will result in error

        @staticmethod
        def from_domain(obj: sqlalchemy_test_model):
            dao = TestModelDao(
                id=obj.id,
                name=obj.name,
                description=obj.description,
            )
            dao.items = [
                sqlalchemy_test_sub_model_dao.from_domain(e) for e in obj.items
            ]
            return dao

        @staticmethod
        def mapper(meta: MetaData, foreign_keys: Dict[str, Mapper]) -> Mapper:
            return mapper(
                TestModelDao,
                TestModelDao.table(meta),
            )

        @staticmethod
        def table(meta: MetaData) -> Table:
            return Table(
                "test_model",
                meta,
                Column("id", String(), primary_key=True),
                Column("name", String()),
                Column("description", String()),
            )

    return TestModelDao


@pytest.fixture
def sqlalchemy_test_sub_model_dao(sqlalchemy_test_sub_model):
    from sqlalchemy import Column, ForeignKey, MetaData, String, Table
    from sqlalchemy.orm import Mapper, mapper, relationship

    from fractal.contrib.sqlalchemy.repositories import SqlAlchemyDao

    @dataclass
    class TestSubModelDao(SqlAlchemyDao):
        id: str
        name: str
        item_id: str

        @staticmethod
        def from_domain(obj: sqlalchemy_test_sub_model):
            return TestSubModelDao(
                id=obj.id,
                name=obj.name,
                item_id=obj.item_id,
            )

        @staticmethod
        def mapper(meta: MetaData, foreign_keys: Dict[str, Mapper]) -> Mapper:
            return mapper(
                TestSubModelDao,
                TestSubModelDao.table(meta),
                properties={
                    k: relationship(v, uselist=False, backref="items")
                    for k, v in foreign_keys.items()
                },
            )

        @staticmethod
        def table(meta: MetaData) -> Table:
            return Table(
                "test_sub_model",
                meta,
                Column("id", String(), primary_key=True),
                Column("name", String()),
                Column("item_id", ForeignKey("test_model.id")),
            )

    return TestSubModelDao


@pytest.fixture
def sqlalchemy_application_mapper(
    sqlalchemy_test_model_dao, sqlalchemy_test_sub_model_dao
):
    from sqlalchemy import MetaData

    from fractal.contrib.sqlalchemy.repositories import DaoMapper

    class ApplicationMapper(DaoMapper):
        def application_mappers(self, meta: MetaData):
            mapper = sqlalchemy_test_model_dao.mapper(meta, {})
            sqlalchemy_test_sub_model_dao.mapper(meta, {"_test_model": mapper})

    return ApplicationMapper


@pytest.fixture
def sqlalchemy_application_mapper_error(
    sqlalchemy_test_model_dao_error, sqlalchemy_test_sub_model_dao
):
    from sqlalchemy import MetaData

    from fractal.contrib.sqlalchemy.repositories import DaoMapper

    class ApplicationMapper(DaoMapper):
        def application_mappers(self, meta: MetaData):
            mapper = sqlalchemy_test_model_dao_error.mapper(meta, {})
            sqlalchemy_test_sub_model_dao.mapper(meta, {"_test_model": mapper})

    return ApplicationMapper


@pytest.fixture
def sqlalchemy_test_repository(
    sqlalchemy_test_model, sqlalchemy_test_model_dao, sqlalchemy_application_mapper
):
    from fractal.contrib.sqlalchemy.repositories import SqlAlchemyRepositoryMixin
    from fractal.core.repositories import Repository

    class TestRepository(Repository[sqlalchemy_test_model], ABC):
        entity = sqlalchemy_test_model

    class SqlAlchemyTestRepository(
        TestRepository,
        SqlAlchemyRepositoryMixin[sqlalchemy_test_model, sqlalchemy_test_model_dao],
    ):
        entity_dao = sqlalchemy_test_model_dao
        application_mapper = sqlalchemy_application_mapper

    return SqlAlchemyTestRepository(connection_str="sqlite://")


@pytest.fixture
def sqlalchemy_test_repository_error(
    sqlalchemy_test_model,
    sqlalchemy_test_model_dao_error,
    sqlalchemy_application_mapper_error,
):
    from fractal.contrib.sqlalchemy.repositories import SqlAlchemyRepositoryMixin
    from fractal.core.repositories import Repository

    class TestRepository(Repository[sqlalchemy_test_model], ABC):
        entity = sqlalchemy_test_model

    class SqlAlchemyTestRepository(
        TestRepository,
        SqlAlchemyRepositoryMixin[
            sqlalchemy_test_model, sqlalchemy_test_model_dao_error
        ],
    ):
        entity_dao = sqlalchemy_test_model_dao_error
        application_mapper = sqlalchemy_application_mapper_error

    return SqlAlchemyTestRepository(connection_str="sqlite://")



import pytest


def test_add_handler(command_bus, command, command_handler):
    command_bus.add_handler(command_handler)

    assert command_bus.handlers[command.__class__.__name__] == [command_handler]


def test_set_handler(command_bus, command, command_handler):
    command_bus.set_handlers([command_handler])

    assert command_bus.handlers[command.__class__.__name__] == [command_handler]


def test_handle(command_bus, command, command_handler):
    command_bus.add_handler(command_handler)

    assert command_bus.handle(command)[command_handler.__class__.__name__] == 1


@pytest.mark.asyncio
async def test_handle_async(command_bus, command, async_command_handler):
    command_bus.add_handler(async_command_handler)

    assert (await command_bus.handle_async(command))[
        async_command_handler.__class__.__name__
    ] == 1



def test_get_payload(token_roles_service_fractal, token):
    from fractal.contrib.fastapi.routers.tokens import get_payload

    inner = get_payload(token_roles_service_fractal)
    inner(token)


def test_get_payload_roles(token_roles_service_fractal, admin_role_token):
    from fractal.contrib.fastapi.routers.tokens import get_payload_roles

    inner = get_payload_roles(token_roles_service_fractal, endpoint="test")
    inner(admin_role_token)



from dataclasses import dataclass
from datetime import date, datetime
from decimal import Decimal
from typing import List

import pytest


@pytest.fixture
def decimal():
    return Decimal(1.1)


@pytest.fixture
def model_instance(now, decimal):
    from fractal.core.models import Model

    @dataclass
    class M(Model):
        string: str = "name"
        number: int = 1
        string_list: List[str] = None
        now: datetime = None
        decimal_number: Decimal = decimal

        def __post_init__(self):
            if not self.string_list:
                self.string_list = ["a", "b"]
            if not self.now:
                self.now = now

    return M()


@pytest.fixture
def nested_model_instance(model_instance):
    from fractal.core.models import Model

    @dataclass
    class M(Model):
        nested_list: List = None

        def __post_init__(self):
            if not self.nested_list:
                self.nested_list = [model_instance]

    return M()


def test_model(model_instance, now, decimal):
    assert model_instance.asdict() == {
        "string": "name",
        "number": 1,
        "string_list": ["a", "b"],
        "now": now.isoformat(),
        "decimal_number": f"{decimal:.2f}",
    }


def test_model_skip_types(model_instance, now, decimal):
    assert model_instance.asdict(skip_types=(list, date, Decimal,)) == {
        "string": "name",
        "number": 1,
        "string_list": ["a", "b"],
        "now": now,
        "decimal_number": decimal,
    }


def test_model_nested(nested_model_instance, now, decimal):
    assert nested_model_instance.asdict() == {
        "nested_list": [
            {
                "string": "name",
                "number": 1,
                "string_list": ["a", "b"],
                "now": now.isoformat(),
                "decimal_number": f"{decimal:.2f}",
            }
        ]
    }


def test_model_nested_skip_types(nested_model_instance, model_instance, now, decimal):
    m = nested_model_instance.asdict(skip_types=(list, date, Decimal,))[
        "nested_list"
    ][0]
    assert type(m) == type(model_instance)
    assert m.now == now


def test_model_nested_skip_types_no_list(nested_model_instance, now, decimal):
    assert nested_model_instance.asdict(skip_types=(date, Decimal,)) == {
        "nested_list": [
            {
                "string": "name",
                "number": 1,
                "string_list": ["a", "b"],
                "now": now,
                "decimal_number": decimal,
            }
        ]
    }



from dataclasses import asdict
from datetime import date
from typing import Iterator, Optional

from fractal_specifications.contrib.google_firestore.specifications import (
    FirestoreSpecificationBuilder,
)
from fractal_specifications.generic.specification import Specification
from google.cloud import firestore
from google.cloud.firestore_v1 import Client

from fractal import Settings
from fractal.contrib.gcp import SettingsMixin
from fractal.core.exceptions import ObjectNotFoundException
from fractal.core.repositories import Entity, Repository
from fractal.core.repositories.sort_repository_mixin import SortRepositoryMixin


def get_firestore_client(settings: Settings):
    if not hasattr(settings, "firestore_client"):
        import firebase_admin
        from firebase_admin import firestore

        cred = None
        if service_account_key := getattr(settings, "GCP_SERVICE_ACCOUNT_KEY"):
            from firebase_admin import credentials

            cred = credentials.Certificate(service_account_key)

        firebase_admin.initialize_app(cred)
        settings.firestore_client = firestore.client()
    return settings.firestore_client


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


class FirestoreRepositoryMixin(SettingsMixin, Repository[Entity]):
    """
    https://github.com/GoogleCloudPlatform/python-docs-samples/blob/46fa5a588858021ea32350584a4ee178cd7c1f33/firestore/cloud-client/snippets.py#L62-L66
    """

    entity = Entity

    def __init__(self, *args, **kwargs):
        super(FirestoreRepositoryMixin, self).__init__(*args, **kwargs)

        client: Client = get_firestore_client(self.settings)
        if app_name := getattr(self.settings, "APP_NAME"):
            self.collection = client.collection(
                f"{app_name.lower()}-{self.entity.__name__.lower()}"
            )
        else:
            self.collection = client.collection(self.entity.__name__.lower())

    def add(self, entity: Entity) -> Entity:
        doc_ref = self.collection.document(entity.id)
        doc_ref.set(asdict(entity))
        return entity

    def update(self, entity: Entity, *, upsert=False) -> Entity:
        if not upsert:
            doc_ref = self.collection.document(entity.id)
            doc = doc_ref.get()
            if doc.exists:
                doc_ref.set(asdict(entity))
            return entity
        return self.add(entity)

    def remove_one(self, specification: Specification):
        entity = self.find_one(specification)
        self.collection.document(entity.id).delete()

    def find_one(self, specification: Specification) -> Optional[Entity]:
        _filter = FirestoreSpecificationBuilder.build(specification)
        collection = self.collection
        if _filter:
            if isinstance(_filter, list):
                for f in _filter:
                    collection = collection.where(*f)
            else:
                collection = collection.where(*_filter)
        for doc in filter(
            lambda i: specification.is_satisfied_by(AttrDict(i.to_dict())),
            collection.stream(),
        ):
            return self.entity.from_dict(doc.to_dict())
        if self.object_not_found_exception:
            raise self.object_not_found_exception
        raise ObjectNotFoundException(f"{self.entity.__name__} not found!")

    def find(self, specification: Specification = None) -> Iterator[Entity]:
        _filter = FirestoreSpecificationBuilder.build(specification)
        collection = self.collection
        if _filter:
            if isinstance(_filter, list):
                for f in _filter:
                    collection = collection.where(*f)
            else:
                collection = collection.where(*_filter)
        for doc in collection.stream():
            yield self.entity.from_dict(doc.to_dict())

    def is_healthy(self) -> bool:
        return True


class FirestoreRepositoryDictMixin(FirestoreRepositoryMixin[Entity]):
    def add(self, entity: Entity) -> Entity:
        doc_ref = self.collection.document(entity.id)
        doc_ref.set(entity.asdict(skip_types=(date,)))
        return entity

    def update(self, entity: Entity, *, upsert=False) -> Entity:
        if not upsert:
            doc_ref = self.collection.document(entity.id)
            doc = doc_ref.get()
            if doc.exists:
                doc_ref.set(entity.asdict(skip_types=(date,)))
            return entity
        return self.add(entity)


class FirestoreSortRepositoryMixin(SortRepositoryMixin[Entity]):
    def find_sort(
        self, specification: Specification = None, *, order_by: str = "", limit: int = 0
    ) -> Iterator[Entity]:
        _filter = FirestoreSpecificationBuilder.build(specification)
        collection = self.collection
        if _filter:
            if isinstance(_filter, list):
                for f in _filter:
                    collection = collection.where(*f)
            else:
                collection = collection.where(*_filter)
        if order_by:
            if reverse := order_by.startswith("-"):
                order_by = order_by[1:]
            collection = collection.order_by(
                order_by,
                direction=firestore.Query.DESCENDING
                if reverse
                else firestore.Query.ASCENDING,
            )
        if limit:
            collection = collection.limit(limit)
        for doc in collection.stream():
            yield self.entity.from_dict(doc.to_dict())



"""Fractal is a scaffolding toolkit for building SOLID logic for your Python applications."""

__version__ = "3.0.0"

from abc import ABC

from fractal.core.utils.application_context import ApplicationContext
from fractal.core.utils.settings import Settings


class FractalException(Exception):
    pass


class Fractal(ABC):
    settings: Settings = None
    context: ApplicationContext = None

    def __init__(self):
        if not self.settings:
            raise FractalException(
                "Fractal service doesn't provide a 'settings' object."
            )
        if not self.context:
            raise FractalException(
                "Fractal service doesn't provide a 'context' object."
            )



import logging


def init_logging(loglevel):
    # create logger
    logger = logging.getLogger("app")
    logger.setLevel(loglevel)

    # create console handler and set level
    ch = logging.StreamHandler()
    ch.setLevel(loglevel)

    # create formatter
    formatter = logging.Formatter("%(levelname)s - %(asctime)s - %(message)s")

    # add formatter to ch
    ch.setFormatter(formatter)

    # add ch to logger
    logger.addHandler(ch)



from typing import Callable, List, Union

from fractal.core.command_bus.command_bus import CommandBus
from fractal.core.event_sourcing.event import (
    EventCommandMapper,
    ReceivingEvent,
    SendingEvent,
)
from fractal.core.event_sourcing.event_projector import EventProjector


class CommandBusProjector(EventProjector):
    def __init__(
        self,
        command_bus_func: Callable[[], CommandBus],
        mappers: List[EventCommandMapper],
    ):
        self.command_bus_func = command_bus_func
        self.mappers = {
            event: mapper for m in mappers for event, mapper in m().mappers().items()
        }

    def project(self, id: str, event: Union[SendingEvent, ReceivingEvent]):
        if isinstance(event, ReceivingEvent):
            self.command_bus_func().handle(event.to_command())
        elif event.__class__ in self.mappers:
            for mapper in self.mappers[event.__class__]:
                self.command_bus_func().handle(mapper(event))



import json
import logging
import threading
from typing import List, Type

from fractal.contrib.rabbitmq.utils import setup_rabbitmq_connection
from fractal.core.command_bus.command_bus import CommandBus
from fractal.core.event_sourcing.event import ReceivingEvent
from fractal.core.event_sourcing.event_bus import EventBus
from fractal.core.event_sourcing.message import Message

logger = logging.getLogger("app")


class RabbitMqEventBus(EventBus):
    def __init__(
        self,
        command_bus: CommandBus,
        host,
        port,
        username,
        password,
        service_name: str,
        event_classes: List[Type[ReceivingEvent]],
        use_thread=False,
    ):
        self.connection, self.channel = setup_rabbitmq_connection(
            host, port, username, password
        )
        for event_class in event_classes:
            logger.info(
                f"Declaring/binding RabbitMq queue: {service_name}_{event_class.__name__}"
            )
            queue = self.channel.queue_declare(  # TODO remove, should be in infra
                queue=f"{service_name}_{event_class.__name__}",
                durable=True,
            )
            self.channel.queue_bind(
                exchange=event_class.__name__,
                queue=queue.method.queue,
            )
            listener = RabbitMqEventBusListener(
                command_bus,
                host,
                port,
                username,
                password,
                event_class,
                queue.method.queue,
            )
            if use_thread:  # TODO should run in separate container
                thread = threading.Thread(target=listener.run)
                thread.setDaemon(True)
                thread.start()
            else:
                listener.run()


class RabbitMqEventBusListener:
    def __init__(
        self,
        command_bus: CommandBus,
        host,
        port,
        username,
        password,
        event_class,
        queue,
    ):
        self.command_bus = command_bus
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.event_class = event_class
        self.queue = queue

    def run(self):
        logger.info(f"Listening to RabbitMq queue: {self.queue}")
        connection, channel = setup_rabbitmq_connection(
            self.host, self.port, self.username, self.password
        )

        def callback(_ch, _method, _properties, body):
            try:
                message = Message(**json.loads(body))
                event = self.event_class(**json.loads(message.data))
                logger.debug("Received event: {}".format(event))
                command = event.to_command()
                if command:
                    self.command_bus.handle(command)
                _ch.basic_ack(_method.delivery_tag)
            except Exception as e:
                logger.exception(e)
                _ch.basic_nack(_method.delivery_tag)

        channel.basic_consume(self.queue, callback)

        try:
            channel.start_consuming()
        finally:
            connection.close()



import os
from io import StringIO
from typing import List, Tuple

from dotenv import load_dotenv


class Settings(object):
    instance = None

    def __new__(cls, dotenv=True, *args, **kwargs):
        if not isinstance(cls.instance, cls):
            cls.instance = object.__new__(cls, *args, **kwargs)
            if dotenv:
                dotenv_path = None
                if root_dir := getattr(cls, "ROOT_DIR", None):
                    dotenv_path = os.path.join(root_dir, ".env")
                load_dotenv(dotenv_path)
            cls.instance.load()
        return cls.instance

    def load(self):
        """Define here your settings."""

    def reload(self, defaults: dict):
        filelike = StringIO("\n".join([f"{k}={v}" for k, v, in defaults.items()]))
        filelike.seek(0)
        load_dotenv(stream=filelike, override=True)
        self.load()
        Settings.instance = self

    def get_parameters(self, parameters: List[str]) -> Tuple:
        for parameter in parameters:
            if not hasattr(self, parameter):
                from fractal import FractalException

                raise FractalException(
                    f"ApplicationContext does not provide '{parameter}'"
                )
        return tuple(getattr(self, p) for p in parameters)



from dataclasses import asdict, dataclass, fields
from datetime import date
from decimal import Decimal
from enum import Enum
from typing import Dict


@dataclass
class Model:
    @classmethod
    def clean(cls, **kwargs):
        field_names = set(f.name for f in fields(cls))
        return cls(**{k: v for k, v in kwargs.items() if k in field_names})  # NOQA

    @classmethod
    def from_dict(cls, data):
        return cls.clean(**data)

    def update(self, model: Dict):
        current = asdict(self)
        current.update(model)
        return self.from_dict(current)

    def asdict(self, *, skip_types=None):
        if skip_types is None:
            skip_types = []
        field_names = set(
            f.name for f in fields(self) if f.name not in self.calculated_fields()
        )

        def _asdict(v):
            if isinstance(v, Model):
                return v.asdict(skip_types=skip_types)
            elif isinstance(v, list) and list not in skip_types:
                return [_asdict(i) for i in v]
            elif isinstance(v, Decimal) and Decimal not in skip_types:
                return f"{v:.2f}"
            elif isinstance(v, date) and date not in skip_types:
                return v.isoformat()
            return v

        ret = {k: _asdict(v) for k, v in self.__dict__.items() if k in field_names}
        return ret

    @staticmethod
    def calculated_fields():
        return []


@dataclass
class Contract(Model):
    pass


@dataclass
class CleanContract(Contract):
    @classmethod
    def clean(cls, **kwargs):
        if "$type" in kwargs:
            kwargs["type"] = kwargs["$type"]
        return super(CleanContract, cls).clean(**kwargs)


class EnumModel(str, Enum):
    @staticmethod
    def values_dict():
        return {str(r).lower().replace("_", "."): r for r in EnumModel}

    @staticmethod
    def from_string(s):
        return EnumModel.values_dict().get(s)



from typing import Generator, Generic, Optional, Type

from fractal_specifications.generic.specification import Specification

from fractal.core.event_sourcing.event_store import EventStore
from fractal.core.event_sourcing.event_stream import EventStream
from fractal.core.event_sourcing.models import EventSourcedAggregateRoot
from fractal.core.exceptions import AggregateRootError
from fractal.core.repositories import Entity, Repository


class EventSourcedRepository(Generic[Entity], Repository[Entity]):
    entity: Type[Entity] = Entity

    def __init__(self, event_store: EventStore):
        self.event_store = event_store

    def commit(self, event_stream: EventStream):
        self.event_store.commit(
            event_stream=event_stream,
            aggregate=self.entity.__name__,
            version=1,
        )

    def add(self, entity: Entity) -> Entity:
        if not isinstance(entity, EventSourcedAggregateRoot):
            raise AggregateRootError
        self.commit(
            EventStream(
                events=entity.release(),
            )
        )
        return entity

    def update(self, entity: Entity, upsert=False) -> Entity:
        return self.add(entity)

    def remove_one(self, specification: Specification):
        raise NotImplementedError

    def find_one(self, specification: Specification) -> Optional[Entity]:
        raise NotImplementedError

    def find(
        self, specification: Specification = None
    ) -> Generator[Entity, None, None]:
        raise NotImplementedError

    def is_healthy(self) -> bool:
        return self.event_store.is_healthy()



def test_print_event_projector(print_event_projector, sending_event, capsys):
    import uuid

    id = str(uuid.uuid4())
    print_event_projector.project(id, sending_event)

    printed = capsys.readouterr().out
    assert sending_event.__class__.__name__ in printed
    assert id in printed



import json
import logging
import threading
from typing import List, Type

from kafka import KafkaConsumer

from fractal.core.command_bus.command_bus import CommandBus
from fractal.core.event_sourcing.event import ReceivingEvent
from fractal.core.event_sourcing.event_bus import EventBus
from fractal.core.event_sourcing.message import Message

logger = logging.getLogger("app")


class KafkaEventBus(EventBus):
    def __init__(
        self,
        command_bus: CommandBus,
        host,
        port,
        username,
        password,
        service_name: str,
        aggregate: str,
        event_classes: List[Type[ReceivingEvent]],
        use_thread=False,
    ):
        listener = KafkaEventBusListener(
            command_bus,
            host,
            port,
            username,
            password,
            service_name,
            aggregate,
            event_classes,
        )
        if use_thread:  # TODO should run in separate container
            thread = threading.Thread(target=listener.run)
            thread.setDaemon(True)
            thread.start()
        else:
            listener.run()


class KafkaEventBusListener:
    def __init__(
        self,
        command_bus: CommandBus,
        host,
        port,
        username,
        password,
        service_name: str,
        aggregate: str,
        event_classes,
    ):
        self.command_bus = command_bus
        self.bootstrap_servers = f"{host}:{port}"
        self.event_classes = {i.__name__: i for i in event_classes}
        self.service_name = service_name
        self.aggregate = aggregate

    def run(self):
        topic = f"{self.service_name}.{self.aggregate}"
        logger.info(f"Listening to RabbitMq queue: {topic}")

        consumer = KafkaConsumer(topic, bootstrap_servers=self.bootstrap_servers)
        for message in consumer:
            logger.info(f"Received message: {message}")
            message = Message(**json.loads(message.value))
            event = self.event_classes[message.event](**json.loads(message.data))
            logger.info("Received event: {}".format(event))
            command = event.to_command()
            if command:
                self.command_bus.handle(command)
